#version 330

layout (std140) uniform Materials
{
	vec4 diffuse;
	vec4 ambient;
	vec4 specular;
	vec4 emissive;
	float shininess;
	int texCount;
};

layout (std140) uniform Lights
{
	vec3 l_dir;		// camera space
};

in Data
{
	vec3 normal;	// camera space
	vec4 eye;		// camera space
	vec4 shadowCoord;
} DataIn;

// TODO 7:
uniform sampler2DShadow shadowUnit;

out vec4 outputF;

//----------------------------------------------------------------
// TODO:
// 6) Display the value in the shadow texture, without doing any comparison (yet)
// 7) Change the sampler type, do the comparison. Still, display the value.
// 8) Compute the lighting propertly with shadows.
//----------------------------------------------------------------

void main()
{
	// TODO 6,7: Get the value from the texture and display it
	float shadow = textureProj(shadowUnit, DataIn.shadowCoord);

	//outputF = vec4(shadow);
	//return;

	// Normalize both input vectors
	vec3 n = normalize(DataIn.normal);
	vec3 e = normalize(vec3(DataIn.eye));

	// Compute diffuse intensity
	float intensity = max(dot(n, l_dir), 0.0f);

	// If the vertex is lit compute the specular color
	vec4 spec = vec4(0.0f);
	if (intensity > 0.0f)
	{
		// Compute the half vector
		vec3 h = normalize(l_dir + e);	
		// Compute the specular term into spec
		float intSpec = max(dot(h, n), 0.0f);
		spec = specular * pow(intSpec, shininess);
	}
	
	// TODO 8: Apply the shadow
	vec4 col = ambient + (intensity * diffuse + spec) * shadow;
	outputF = clamp(col, 0.0f, 1.0f);
}