//
// Lighthouse3D.com
//
//
// This demo was built for learning purposes only.
// Some code could be severely optimised, but I tried to
// keep as simple and clear as possible.
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//
// If you do use it I would love to hear about it. Just post a comment
// at Lighthouse3D.com

// Have Fun :-)

//
// Program for lecturing purposes by Marek Vinkler (xvinkl@fi.muni.cz).
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//

#include <cmath>

// Include GLEW to access OpenGL 3.3 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

// Use Very Simple Libs
#include "vsMathLib.h"
#include "vsShaderLib.h"
#include "vsResSurfRevLib.h"

//----------------------------------------------------------------
// TODO:
// 1) Inspect functions "setupShaders", "init", "initVSL", "changeSize" and "renderScene".
// 2) Temporarily override projection and view matrices to look from the point of the light.
// 3) Revert back the projection and view matrices.
// 4) Create the shadow texture and framebuffer, too bad we cannot see the result :-(
// 5) Render into the shadow texture, don't forget the TODOs in the shader, too bad we cannot see the result :-(
// 6) Use the shadow texture and display it, don't forget the TODOs in the shader
// 7) Use comparision, don't forget the TODOs in the shader
// 8) Compute proper illumination (shader only)
// 9) Apply some techniques to remove the artefacts
// 10) Try changing the interpolation to remove pixelated edges
// 11) Try increasing the shadow texture resolution to remove pixelated edges
// 12) Do some experiments:
//		- Use ortographic projection matrix - we are using directional light.
//		- Try a different animation
//----------------------------------------------------------------

VSMathLib *vsml;
VSShaderLib shader, shaderShadow;
VSResSurfRevLib torusCaster, sphereReceiver;

// Current window size
int win_width, win_height;

// Size of the shadow texture
// TODO 11: higher means better quality
int shadowMapSize = 1024;

// Shadow texture object
GLuint shadowMap;

// Framebuffer with texture object
GLuint shadowFramebuffer;

// Projection matrix for shadow generation
float shadowProjectionMatrix[16];
// View matrix for shadow generation
float shadowViewMatrix[16];

// Matrix for remapping [-1,1] clip coordinates into [0,1] texture coordinates
float remapMatrix[16] = {
	0.5f, 0.0f, 0.0f, 0.0f,
	0.0f, 0.5f, 0.0f, 0.0f,
	0.0f, 0.0f, 0.5f, 0.0f,
	0.5f, 0.5f, 0.5f, 1.0f};

// Matrix for camera projection
float cameraProjectionMatrix[16];

// Light direction
float lightDir[4] = {1.0f, 1.0f, 1.0f, 0.0f};

// Distance of the light
float lightDist = 3.f;

// Camera Position
float camX, camY, camZ;

// Mouse Tracking Variables
int startX, startY, tracking = 0;

// Camera Spherical Coordinates
float alpha = 0.0f, beta = 0.0f;
float r = 5.0f;

// Camera offset
float offsetX = 0.f, offsetY = 0.f;

// Time counter
float time = 0.f;

// ------------------------------------------------------------
//
// Reshape Callback Function
//

void changeSize(int w, int h)
{
	// Prevent a divide by zero, when window is too short
	if(h == 0)
		h = 1;

	win_width = w;
	win_height = h;

	// Update the projection matrix of the camera
	vsml->loadIdentity(VSMathLib::PROJECTION);
	vsml->perspective(53.13f, float(win_width) / float(win_height), 0.1f, 10000.0f);
	memcpy(cameraProjectionMatrix, vsml->get(VSMathLib::PROJECTION), sizeof(cameraProjectionMatrix)); // Save to global variable
}


// ------------------------------------------------------------
//
// Render stuff
//

void renderTorus()
{
	// Move the torus so it casts a nice shadow
	// TODO 12: Try a different animation
	vsml->pushMatrix(VSMathLib::MODEL);
	vsml->translate(0.0f, 1.0f, 0.0f);
	vsml->rotate(time, 1.0f, 1.0f, 1.0f);
	torusCaster.render();
	vsml->popMatrix(VSMathLib::MODEL);
}

void renderSphere()
{
	// Move the sphere so that it gets the shadow
	vsml->pushMatrix(VSMathLib::MODEL);
	vsml->translate(-5.0f, -2.5f, -10.0f);
	sphereReceiver.render();
	vsml->popMatrix(VSMathLib::MODEL);
}

void renderScene()
{
	//-----------------------------
	//--  Shadow map generation  --

	// TODO 5: Bind the framebuffer with the shadow texture
	glBindFramebuffer(GL_FRAMEBUFFER, shadowFramebuffer);

	// TODO 5: Set the viewport to the texture size
	glViewport(0, 0, shadowMapSize, shadowMapSize);

	// TODO 5: Disable writing of color information, or not?
	//glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

	// TODO 5: Clear the depth buffer
	glClear(GL_DEPTH_BUFFER_BIT);

	// TODO 5: Use the shader for shadow map generation
	glUseProgram(shaderShadow.getProgramIndex());

	// TODO 9: Render offseted back faces
	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(1.0f, 0.0f);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_FRONT);

	// TODO 5: Load the projection matrix
	vsml->loadMatrix(VSMathLib::PROJECTION, shadowProjectionMatrix);

	// TODO 5: Load the view matrix
	vsml->loadMatrix(VSMathLib::VIEW, shadowViewMatrix);
	
	// TODO 5: Render the scene
	vsml->loadIdentity(VSMathLib::MODEL);
	renderTorus();
	renderSphere();

	// TODO 9: Revert (some) states
	glDisable(GL_POLYGON_OFFSET_FILL);
	glDisable(GL_CULL_FACE);

	//-------------------
	//--  Final image  --

	// Bind the window framebuffer
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	
	// TODO 5: Enable writing of color information if disabled
	//glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

	// Set the viewport to the window size
	glViewport(0, 0, win_width, win_height);

	// Clear the buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Use shader for rendering with shadow
	glUseProgram(shader.getProgramIndex());

	// TODO 6: Bind the generated shadow map
	glBindTexture(GL_TEXTURE_2D, shadowMap);

	// Load the projection matrix
	vsml->loadMatrix(VSMathLib::PROJECTION, cameraProjectionMatrix);
	// TODO 2,3: Override the projection matrix

	// Compute and set the view matrix
	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->lookAt(camX, camY, camZ, 0,0,0, 0,1,0);
	vsml->translate(VSMathLib::VIEW, offsetX, offsetY, 0.f);
	// TODO 2,3: Override the view matrix

	// Transform light to camera space and send it to GLSL
	float res[4];
	vsml->multMatrixPoint(VSMathLib::VIEW, lightDir, res);
	vsml->normalize(res);
	shader.setBlockUniform("Lights", "l_dir", res);

	// Compute and set the shadow matrix
	vsml->loadIdentity(VSMathLib::AUX0);
	vsml->multMatrix(VSMathLib::AUX0, remapMatrix);
	vsml->multMatrix(VSMathLib::AUX0, shadowProjectionMatrix);
	vsml->multMatrix(VSMathLib::AUX0, shadowViewMatrix);

	shader.setUniform("time", time);

	// Render the scene
	vsml->loadIdentity(VSMathLib::MODEL);
	renderTorus();
	renderSphere();

	// Move the time forward
	time += 1.0f;

	// TODO 6: Unbind the 'shadowMap' texture. It is not good for it to be bound when we render into it.
	glBindTexture(GL_TEXTURE_2D, 0);

	// Swap buffers
	glutSwapBuffers();
}


// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key)
	{
		case 27:
			glutLeaveMainLoop();
			break;

		case 'c': printf("Camera Spherical Coordinates (%f, %f, %f)\n", alpha, beta, r);
			break;

		case 'm': glEnable(GL_MULTISAMPLE);
			break;

		case 'n': glDisable(GL_MULTISAMPLE);
			 break;
	}

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// ------------------------------------------------------------
//
// Mouse Events
//

void processMouseButtons(int button, int state, int xx, int yy)
{
	// Start tracking the mouse
	if (state == GLUT_DOWN)
	{
		startX = xx;
		startY = yy;
		if (button == GLUT_LEFT_BUTTON)
			tracking = 1;
		else if (button == GLUT_RIGHT_BUTTON)
			tracking = 2;
		else if (button == GLUT_MIDDLE_BUTTON)
			tracking = 3;
	}

	// Stop tracking the mouse
	else if (state == GLUT_UP)
	{
		if (tracking == 1)
		{
			alpha -= (xx - startX);
			beta += (yy - startY);
			if (beta > 85.0f)
				beta = 85.0f;
			else if (beta < -85.0f)
				beta = -85.0f;
		}
		else if (tracking == 2)
		{
			r += (yy - startY) * 0.01f;
			if (r < 0.1f)
				r = 0.1f;
		}
		tracking = 0;
	}
}

// Track mouse motion while buttons are pressed
void processMouseMotion(int xx, int yy)
{
	int deltaX, deltaY;
	float alphaAux, betaAux;
	float rAux;

	deltaX =  - xx + startX;
	deltaY =    yy - startY;

	// Left mouse button: move camera
	if (tracking == 1)
	{
		alphaAux = alpha + deltaX;
		betaAux = beta + deltaY;

		if (betaAux > 85.0f)
			betaAux = 85.0f;
		else if (betaAux < -85.0f)
			betaAux = -85.0f;
		rAux = r;
	}
	// Right mouse button: zoom
	else if (tracking == 2)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r + (deltaY * 0.01f);
		if (rAux < 0.1f)
			rAux = 0.1f;
	}
	// Middle mouse button: move object
	else if (tracking == 3)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r;

		offsetX -= deltaX * 0.0001f;
		offsetY -= deltaY * 0.0001f;
	}

	camX = rAux * sin(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camZ = rAux * cos(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camY = rAux *   						       sin(betaAux * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}

void mouseWheel(int wheel, int direction, int x, int y)
{
	r += direction * 0.1f;
	if (r < 0.1f)
		r = 0.1f;

	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// --------------------------------------------------------
//
// Shader Stuff
//

bool setupShaders()
{
	// Shader for shadow usage
	shader.init();
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/useShadow.vert");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/useShadow.frag");

	// Set semantics for the shader variables
	shader.setProgramOutput(0, "outputF");
	shader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	shader.setVertexAttribName(VSShaderLib::NORMAL_ATTRIB, "normal");

	shader.prepareProgram();
	printf("InfoLog for scene shader\n%s\n\n", shader.getAllInfoLogs().c_str());

	shader.setUniform("shadowUnit", 0); // Add sampler uniform

	// Shader for shadow generation
	shaderShadow.init();
	shaderShadow.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/genShadow.vert");
	shaderShadow.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/genShadow.frag");
	
	// Set semantics for the shader variables
	shaderShadow.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");

	shaderShadow.prepareProgram();
	printf("InfoLog for shadow shader\n%s\n\n", shaderShadow.getAllInfoLogs().c_str());

	return(shader.isProgramLinked() && shaderShadow.isProgramLinked());
}


// ------------------------------------------------------------
//
// Model loading and OpenGL setup
//

bool init()
{
	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

	torusCaster.createTorus(0.5f, 1.0f, 50, 50);
	torusCaster.setMaterialBlockName("Materials");

	float tD[4] = {0.8f, 0.6f, 0.4f, 1.0f};
	torusCaster.setColor(VSResourceLib::DIFFUSE, tD);
	float tA[4] = {0.2f, 0.15f, 0.1f, 1.0f};
	torusCaster.setColor(VSResourceLib::AMBIENT, tA);

	sphereReceiver.createSphere(10.f, 50);
	sphereReceiver.setMaterialBlockName("Materials");

	float sD[4] = {0.8f, 0.9f, 0.8f, 1.0f};
	sphereReceiver.setColor(VSResourceLib::DIFFUSE, sD);
	float sA[4] = {0.2f, 0.2f, 0.2f, 1.0f};
	sphereReceiver.setColor(VSResourceLib::AMBIENT, sA);

	// ------------------------------------------------------------
	// Create textures for shadow map
	
	// TODO 4: Create a shadow texture object
	glGenTextures(1, &shadowMap);
	glBindTexture(GL_TEXTURE_2D, shadowMap);

	// TODO 4: Allocate the data for the shadow texture
	glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, shadowMapSize, shadowMapSize, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, 0);

	// TODO 4: Set interpolation and address mode
	// TODO 10: Change the interpolation
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	// TODO 7: Set depth texture comparison function - lower value than texture means 1.0, higher means 0.0
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_REF_TO_TEXTURE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LESS);

	// TODO 4: Unbind, the texture will be bound during rendering
	glBindTexture(GL_TEXTURE_2D, 0);

	// TODO 4: Create framebuffers for shadow textures
	glGenFramebuffers(1, &shadowFramebuffer);
	glBindFramebuffer(GL_FRAMEBUFFER, shadowFramebuffer);

	// TODO 4: Attach the depth texture.
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, shadowMap, 0);

	// TODO 4: Check correctness of shadow framebuffer
	if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
	//if (false)
	{
		printf("GL_FRAMEBUFFER_COMPLETE error 0x%x", glCheckFramebufferStatus(GL_FRAMEBUFFER));

		return false;
	}

	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	// ------------------------------------------------------------

	// Some GL settings
	glEnable(GL_DEPTH_TEST);

	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	glClearDepth(1.0f);

	//----------------------------------------------------------
	// Compute the matrices for rendering into the shadow map

	// Compute the projection matrix
	// TODO 12: Try ortho projection
	vsml->loadIdentity(VSMathLib::PROJECTION);
	vsml->perspective(53.13f, 1.0f, 2.0f, 20.0f);
	//vsml->ortho(-2.f, 2.f, -2.f, 2.f, 2.0f, 20.0f);
	memcpy(shadowProjectionMatrix, vsml->get(VSMathLib::PROJECTION), sizeof(shadowProjectionMatrix)); // Save to global variable

	// Compute the view matrix
	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->lookAt(lightDir[0]*lightDist, lightDir[1]*lightDist, lightDir[2]*lightDist, 0,0,0, 0,1,0);
	memcpy(shadowViewMatrix, vsml->get(VSMathLib::VIEW), sizeof(shadowViewMatrix)); // Save to global variable
	
	return true;
}

void initVSL()
{
	// Init VSML
	vsml = VSMathLib::getInstance();
	vsml->setUniformBlockName("Matrices");
	vsml->setUniformName(VSMathLib::PROJ_VIEW_MODEL, "m_pvm");
	vsml->setUniformName(VSMathLib::VIEW, "m_view");
	vsml->setUniformName(VSMathLib::MODEL, "m_model");
	vsml->setUniformName(VSMathLib::NORMAL, "m_normal");
	vsml->setUniformName(VSMathLib::AUX0, "m_shadow"); // from world space to light's texture space
}


// ------------------------------------------------------------
//
// Main function
//

int main(int argc, char **argv)
{
//  GLUT initialization
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGBA|GLUT_MULTISAMPLE);

	glutInitContextVersion(3, 3);
	glutInitContextProfile(GLUT_CORE_PROFILE);
	glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(512,512);
	glutCreateWindow("PV227 - Shadow Mapping");


//  Callback Registration
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutIdleFunc(renderScene);

//	Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);
	glutMouseFunc(processMouseButtons);
	glutMotionFunc(processMouseMotion);

	glutMouseWheelFunc(mouseWheel);

//	Return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

//	Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("Version: %s\n", glGetString(GL_VERSION));
	printf("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));
	int param;
	glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &param);
	if(param == GL_CONTEXT_CORE_PROFILE_BIT)
		printf("Context Profile: Core\n");
	else
		printf("Context Profile: Compatibility\n");

	if(!setupShaders())
		exit(1);

	initVSL();

	if(!init())
		exit(1);

//  GLUT main loop
	glutMainLoop();

	return(0);
}
